import json
from enum import Enum
from inspect import signature
from typing import Any, Callable, Iterator, List, Mapping, Optional, Tuple

from crewai import Agent
from crewai.tools.tool_calling import ToolCalling
from fi_instrumentation import get_attributes_from_context, safe_json_dumps
from fi_instrumentation.fi_types import FiSpanKindValues, SpanAttributes
from opentelemetry import context as context_api
from opentelemetry import trace as trace_api
from opentelemetry.util.types import AttributeValue


class SafeJSONEncoder(json.JSONEncoder):
    """
    Safely encodes non-JSON-serializable objects.
    """

    def default(self, o: Any) -> Any:
        try:
            return super().default(o)
        except TypeError:
            if hasattr(o, "dict") and callable(
                o.dict
            ):  # pydantic v1 models, e.g., from Cohere
                return o.dict()
            return repr(o)


def _flatten(
    mapping: Optional[Mapping[str, Any]]
) -> Iterator[Tuple[str, AttributeValue]]:
    if not mapping:
        return
    for key, value in mapping.items():
        if value is None:
            continue
        if isinstance(value, Mapping):
            for sub_key, sub_value in _flatten(value):
                yield f"{key}.{sub_key}", sub_value
        elif isinstance(value, List) and any(
            isinstance(item, Mapping) for item in value
        ):
            for index, sub_mapping in enumerate(value):
                for sub_key, sub_value in _flatten(sub_mapping):
                    yield f"{key}.{index}.{sub_key}", sub_value
        else:
            if isinstance(value, Enum):
                value = value.value
            yield key, value


def _get_input_value(method: Callable[..., Any], *args: Any, **kwargs: Any) -> str:
    """
    Parses a method call's inputs into a JSON string. Ensures a consistent
    output regardless of whether the those inputs are passed as positional or
    keyword arguments.
    """

    # For typical class methods, the corresponding instance of inspect.Signature
    # does not include the self parameter. However, the inspect.Signature
    # instance for __call__ does include the self parameter.

    tool = kwargs.get("calling")
    if tool and isinstance(tool, ToolCalling):
        return safe_json_dumps(
            {
                "tool_name": tool.tool_name,
                "tool_args": tool.arguments,
            },
            cls=SafeJSONEncoder,
        )

    if args and args[0] and isinstance(args[0], Agent):
        return safe_json_dumps(
            {
                "role": args[0].role,
                "goal": args[0].goal,
                "backstory": args[0].backstory,
            },
            cls=SafeJSONEncoder,
        )

    method_signature = signature(method)
    first_parameter_name = next(iter(method_signature.parameters), None)
    signature_contains_self_parameter = first_parameter_name in ["self"]
    bound_arguments = method_signature.bind(
        *(
            [
                None
            ]  # the value bound to the method's self argument is discarded below, so pass None
            if signature_contains_self_parameter
            else []  # no self parameter, so no need to pass a value
        ),
        *args,
        **kwargs,
    )
    return safe_json_dumps(
        {
            **{
                argument_name: argument_value
                for argument_name, argument_value in bound_arguments.arguments.items()
                if argument_name not in ["self", "kwargs"]
            },
            **bound_arguments.arguments.get("kwargs", {}),
        },
        cls=SafeJSONEncoder,
    )


class _ExecuteCoreWrapper:
    def __init__(self, tracer: trace_api.Tracer) -> None:
        self._tracer = tracer

    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: Mapping[str, Any],
    ) -> Any:
        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
            return wrapped(*args, **kwargs)
        if instance:
            span_name = f"{instance.__class__.__name__}.{wrapped.__name__}"
        else:
            span_name = wrapped.__name__
        with self._tracer.start_as_current_span(
            span_name,
            attributes=dict(
                _flatten(
                    {
                        FI_SPAN_KIND: FiSpanKindValues.AGENT,
                        SpanAttributes.INPUT_VALUE: _get_input_value(
                            wrapped,
                            *args,
                            **kwargs,
                        ),
                    }
                )
            ),
            record_exception=False,
            set_status_on_exception=False,
        ) as span:
            agent = args[0] if args else None
            crew = agent.crew if agent else None
            task = instance

            span.set_attribute(
                SpanAttributes.RAW_INPUT,
                safe_json_dumps(_prepare_args_kwargs(args=args, **kwargs)),
            )

            if crew:
                span.set_attribute("crew_key", crew.key)
                span.set_attribute("crew_id", str(crew.id))
            span.set_attribute("task_key", task.key)
            span.set_attribute("task_id", str(task.id))

            if crew and crew.share_crew:
                span.set_attribute("formatted_description", task.description)
                span.set_attribute("formatted_expected_output", task.expected_output)
            try:
                response = wrapped(*args, **kwargs)
            except Exception as exception:
                span.set_status(
                    trace_api.Status(trace_api.StatusCode.ERROR, str(exception))
                )
                span.record_exception(exception)
                raise

            span.set_status(trace_api.StatusCode.OK)
            span.set_attribute(OUTPUT_VALUE, safe_json_dumps(response))
            span.set_attributes(dict(get_attributes_from_context()))
        return response


class _KickoffWrapper:
    def __init__(self, tracer: trace_api.Tracer) -> None:
        self._tracer = tracer

    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: Mapping[str, Any],
    ) -> Any:
        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
            return wrapped(*args, **kwargs)
        span_name = f"{instance.__class__.__name__}.kickoff"
        with self._tracer.start_as_current_span(
            span_name,
            record_exception=False,
            set_status_on_exception=False,
            attributes=dict(
                _flatten(
                    {
                        FI_SPAN_KIND: FiSpanKindValues.CHAIN,
                    }
                )
            ),
        ) as span:
            crew = instance
            inputs = kwargs.get("inputs", None) or (args[0] if args else None)
            span.set_attribute("crew_key", crew.key)
            span.set_attribute("crew_id", str(crew.id))
            span.set_attribute("crew_inputs", json.dumps(inputs) if inputs else "")
            span.set_attribute(
                SpanAttributes.RAW_INPUT,
                safe_json_dumps(_prepare_args_kwargs(args=args, **kwargs)),
            )
            span.set_attribute(
                "crew_agents",
                json.dumps(
                    [
                        {
                            "key": agent.key,
                            "id": str(agent.id),
                            "role": agent.role,
                            "goal": agent.goal,
                            "backstory": agent.backstory,
                            "verbose?": agent.verbose,
                            "max_iter": agent.max_iter,
                            "max_rpm": agent.max_rpm,
                            "i18n": agent.i18n.prompt_file,
                            "delegation_enabled": agent.allow_delegation,
                            "tools_names": [
                                tool.name.casefold() for tool in agent.tools or []
                            ],
                        }
                        for agent in crew.agents
                    ]
                ),
            )
            span.set_attribute(
                "crew_tasks",
                json.dumps(
                    [
                        {
                            "id": str(task.id),
                            "description": task.description,
                            "expected_output": task.expected_output,
                            "async_execution?": task.async_execution,
                            "human_input?": task.human_input,
                            "agent_role": task.agent.role if task.agent else "None",
                            "agent_key": task.agent.key if task.agent else None,
                            "context": (
                                [task.description for task in task.context]
                                if task.context
                                else None
                            ),
                            "tools_names": [
                                tool.name.casefold() for tool in task.tools or []
                            ],
                        }
                        for task in crew.tasks
                    ]
                ),
            )
            try:
                crew_output = wrapped(*args, **kwargs)
                usage_metrics = crew.usage_metrics
                if isinstance(usage_metrics, dict):
                    if (
                        prompt_tokens := usage_metrics.get("prompt_tokens")
                    ) is not None:
                        span.set_attribute(LLM_TOKEN_COUNT_PROMPT, int(prompt_tokens))
                    if (
                        completion_tokens := usage_metrics.get("completion_tokens")
                    ) is not None:
                        span.set_attribute(
                            LLM_TOKEN_COUNT_COMPLETION, int(completion_tokens)
                        )
                    if (total_tokens := usage_metrics.get("total_tokens")) is not None:
                        span.set_attribute(LLM_TOKEN_COUNT_TOTAL, int(total_tokens))
                else:
                    # version 0.51 and onwards
                    span.set_attribute(
                        LLM_TOKEN_COUNT_PROMPT, usage_metrics.prompt_tokens
                    )
                    span.set_attribute(
                        LLM_TOKEN_COUNT_COMPLETION, usage_metrics.completion_tokens
                    )
                    span.set_attribute(
                        LLM_TOKEN_COUNT_TOTAL, usage_metrics.total_tokens
                    )

            except Exception as exception:
                span.set_status(
                    trace_api.Status(trace_api.StatusCode.ERROR, str(exception))
                )
                span.record_exception(exception)
                raise
            span.set_status(trace_api.StatusCode.OK)
            if crew_output_dict := crew_output.to_dict():

                span.set_attribute(OUTPUT_VALUE, json.dumps(crew_output_dict))
                span.set_attribute(OUTPUT_MIME_TYPE, "application/json")
            else:
                span.set_attribute(OUTPUT_VALUE, str(crew_output))
            span.set_attributes(dict(get_attributes_from_context()))
        return crew_output


class _ToolUseWrapper:
    def __init__(self, tracer: trace_api.Tracer) -> None:
        self._tracer = tracer

    def __call__(
        self,
        wrapped: Callable[..., Any],
        instance: Any,
        args: Tuple[Any, ...],
        kwargs: Mapping[str, Any],
    ) -> Any:
        if context_api.get_value(context_api._SUPPRESS_INSTRUMENTATION_KEY):
            return wrapped(*args, **kwargs)
        if instance:
            span_name = f"{instance.__class__.__name__}.{wrapped.__name__}"
        else:
            span_name = wrapped.__name__
        with self._tracer.start_as_current_span(
            span_name,
            attributes=dict(
                _flatten(
                    {
                        FI_SPAN_KIND: FiSpanKindValues.TOOL,
                        SpanAttributes.INPUT_VALUE: _get_input_value(
                            wrapped,
                            *args,
                            **kwargs,
                        ),
                    }
                )
            ),
            record_exception=False,
            set_status_on_exception=False,
        ) as span:
            tool = kwargs.get("tool")
            tool_name = ""
            if tool:
                tool_name = tool.name
            span.set_attribute("function_calling_llm", instance.function_calling_llm)
            span.set_attribute(SpanAttributes.TOOL_NAME, tool_name)

            span.set_attribute(
                SpanAttributes.RAW_INPUT,
                safe_json_dumps(_prepare_args_kwargs(args=args, **kwargs)),
            )

            try:
                response = wrapped(*args, **kwargs)
            except Exception as exception:
                span.set_status(
                    trace_api.Status(trace_api.StatusCode.ERROR, str(exception))
                )
                span.record_exception(exception)
                raise
            span.set_status(trace_api.StatusCode.OK)
            span.set_attribute(RAW_OUTPUT, safe_json_dumps(response))
            span.set_attribute(OUTPUT_VALUE, safe_json_dumps(response))
            span.set_attributes(dict(get_attributes_from_context()))
        return response


def _convert_to_dict(obj: Any) -> Any:
    """
    Recursively converts objects to dicts if they have a to_dict method or __dict__ attribute.
    Handles nested lists and dictionaries.
    """
    if hasattr(obj, "to_dict") and callable(obj.to_dict):
        return obj.to_dict()
    elif hasattr(obj, "__dict__"):
        return obj.__dict__
    elif isinstance(obj, dict):
        return {key: _convert_to_dict(value) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [_convert_to_dict(item) for item in obj]
    return obj


def _prepare_args_kwargs(args: Tuple[Any, ...], **kwargs: Any) -> dict:
    """
    Prepares args and kwargs for JSON serialization by converting objects with to_dict
    method to dictionaries. Maintains the original argument structure.
    """
    processed_args = _convert_to_dict(args)
    processed_kwargs = _convert_to_dict(kwargs)
    return {
        "args": processed_args,
        **processed_kwargs,  # Unpacks the processed kwargs directly
    }


INPUT_VALUE = SpanAttributes.INPUT_VALUE
FI_SPAN_KIND = SpanAttributes.FI_SPAN_KIND
OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE
RAW_INPUT = SpanAttributes.RAW_INPUT
RAW_OUTPUT = SpanAttributes.RAW_OUTPUT
OUTPUT_MIME_TYPE = SpanAttributes.OUTPUT_MIME_TYPE
LLM_TOKEN_COUNT_PROMPT = SpanAttributes.LLM_TOKEN_COUNT_PROMPT
LLM_TOKEN_COUNT_COMPLETION = SpanAttributes.LLM_TOKEN_COUNT_COMPLETION
LLM_TOKEN_COUNT_TOTAL = SpanAttributes.LLM_TOKEN_COUNT_TOTAL
